package com.sun.org.apache.bcel.internal.generic;

import java.util.HashMap;
import java.util.Map;
import java.util.Stack;
import com.sun.org.apache.bcel.internal.classfile.ParameterAnnotationEntry;
import com.sun.org.apache.bcel.internal.classfile.RuntimeVisibleParameterAnnotations;
import com.sun.org.apache.bcel.internal.classfile.ParameterAnnotations;
import com.sun.org.apache.bcel.internal.classfile.Utility;
import java.util.Iterator;
import com.sun.org.apache.bcel.internal.classfile.LocalVariable;
import java.util.Arrays;
import java.util.Comparator;
import com.sun.org.apache.bcel.internal.classfile.LineNumber;
import com.sun.org.apache.bcel.internal.classfile.CodeException;
import com.sun.org.apache.bcel.internal.classfile.AnnotationEntry;
import com.sun.org.apache.bcel.internal.classfile.LocalVariableTable;
import com.sun.org.apache.bcel.internal.classfile.LineNumberTable;
import com.sun.org.apache.bcel.internal.classfile.Annotations;
import com.sun.org.apache.bcel.internal.classfile.ExceptionTable;
import com.sun.org.apache.bcel.internal.classfile.Code;
import com.sun.org.apache.bcel.internal.classfile.Method;
import java.util.ArrayList;
import com.sun.org.apache.bcel.internal.util.BCELComparator;
import com.sun.org.apache.bcel.internal.classfile.Attribute;
import java.util.List;
import com.sun.org.apache.bcel.internal.classfile.LocalVariableTypeTable;

public class MethodGen extends FieldGenOrMethodGen
{
    private String class_name;
    private Type[] arg_types;
    private String[] arg_names;
    private int max_locals;
    private int max_stack;
    private InstructionList il;
    private boolean strip_attributes;
    private LocalVariableTypeTable local_variable_type_table;
    private final List<LocalVariableGen> variable_vec;
    private final List<LineNumberGen> line_number_vec;
    private final List<CodeExceptionGen> exception_vec;
    private final List<String> throws_vec;
    private final List<Attribute> code_attrs_vec;
    private List<AnnotationEntryGen>[] param_annotations;
    private boolean hasParameterAnnotations;
    private boolean haveUnpackedParameterAnnotations;
    private static BCELComparator bcelComparator;
    private List<MethodObserver> observers;
    
    public MethodGen(int access_flags, Type return_type, Type[] arg_types, String[] arg_names, String method_name, String class_name, InstructionList il, ConstantPoolGen cp) {
        super(access_flags);
        this.local_variable_type_table = null;
        this.variable_vec = new ArrayList();
        this.line_number_vec = new ArrayList();
        this.exception_vec = new ArrayList();
        this.throws_vec = new ArrayList();
        this.code_attrs_vec = new ArrayList();
        this.hasParameterAnnotations = false;
        this.haveUnpackedParameterAnnotations = false;
        this.setType(return_type);
        this.setArgumentTypes(arg_types);
        this.setArgumentNames(arg_names);
        this.setName(method_name);
        this.setClassName(class_name);
        this.setInstructionList(il);
        this.setConstantPool(cp);
        boolean abstract_ = this.isAbstract() || this.isNative();
        InstructionHandle start = null;
        InstructionHandle end = null;
        if (!abstract_) {
            start = il.getStart();
            if (!this.isStatic() && class_name != null) {
                this.addLocalVariable("this", ObjectType.getInstance(class_name), start, end);
            }
        }
        if (arg_types != null) {
            int size = arg_types.length;
            int length = arg_types.length;
            int j = 0;
            while (j < length) {
                Type arg_type = arg_types[j];
                if (Type.VOID == arg_type) {
                    throw new ClassGenException("'void' is an illegal argument type for a method");
                }
                ++j;
            }
            if (arg_names != null) {
                if (size != arg_names.length) {
                    throw new ClassGenException("Mismatch in argument array lengths: " + size + " vs. " + arg_names.length);
                }
            }
            else {
                arg_names = new String[size];
                for (int i = 0; i < size; ++i) {
                    arg_names[i] = "arg" + i;
                }
                this.setArgumentNames(arg_names);
            }
            if (!abstract_) {
                for (int i = 0; i < size; ++i) {
                    this.addLocalVariable(arg_names[i], arg_types[i], start, end);
                }
            }
        }
    }
    
    public MethodGen(Method m, String class_name, ConstantPoolGen cp) {
        this(m.getAccessFlags(), Type.getReturnType(m.getSignature()), Type.getArgumentTypes(m.getSignature()), null, m.getName(), class_name, ((m.getAccessFlags() & 0x500) == 0x0) ? new InstructionList(m.getCode().getCode()) : null, cp);
        Attribute[] attributes;
        for (Attribute a : attributes = m.getAttributes()) {
            Attribute attribute = a;
            if ((a instanceof Code)) {
                Code c = (Code)a;
                this.setMaxStack(c.getMaxStack());
                this.setMaxLocals(c.getMaxLocals());
                CodeException[] ces = c.getExceptionTable();
                if (ces != null) {
                    for (CodeException ce : ces) {
                        int type = ce.getCatchType();
                        ObjectType c_type = null;
                        if (type > 0) {
                            String cen = m.getConstantPool().getConstantString(type, (byte)7);
                            c_type = ObjectType.getInstance(cen);
                        }
                        int end_pc = ce.getEndPC();
                        int length = m.getCode().getCode().length;
                        InstructionHandle end;
                        if (length == end_pc) {
                            end = this.il.getEnd();
                        }
                        else {
                            end = this.il.findHandle(end_pc);
                            end = end.getPrev();
                        }
                        this.addExceptionHandler(this.il.findHandle(ce.getStartPC()), end, this.il.findHandle(ce.getHandlerPC()), c_type);
                    }
                }
                Attribute[] c_attributes = c.getAttributes();
                for (int length6 = c_attributes.length, n2 = 0; n2 < length6; ++n2) {
                    Attribute c_attribute = a = c_attributes[n2];
                    if ((a instanceof LineNumberTable)) {
                        LineNumber[] ln;
                        for (LineNumber l : ln = ((LineNumberTable)a).getLineNumberTable()) {
                            InstructionHandle ih = this.il.findHandle(l.getStartPC());
                            if (ih != null) {
                                this.addLineNumber(ih, l.getLineNumber());
                            }
                        }
                    }
                    else if ((a instanceof LocalVariableTable)) {
                        this.updateLocalVariableTable((LocalVariableTable)a);
                    }
                    else if ((a instanceof LocalVariableTypeTable)) {
                        this.local_variable_type_table = (LocalVariableTypeTable)a.copy(cp.getConstantPool());
                    }
                    else {
                        this.addCodeAttribute(a);
                    }
                }
            }
            else if ((a instanceof ExceptionTable)) {
                String[] names;
                for (String name2 : names = ((ExceptionTable)a).getExceptionNames()) {
                    this.addException(name2);
                }
            }
            else if ((a instanceof Annotations)) {
                Annotations runtimeAnnotations = (Annotations)a;
                AnnotationEntry[] aes;
                for (AnnotationEntry element : aes = runtimeAnnotations.getAnnotationEntries()) {
                    this.addAnnotationEntry(new AnnotationEntryGen(element, cp, false));
                }
            }
            else {
                this.addAttribute(a);
            }
        }
    }
    
    public LocalVariableGen addLocalVariable(String name, Type type, int slot, InstructionHandle start, InstructionHandle end, int orig_index) {
        byte t = type.getType();
        if (t != 16) {
            int add = type.getSize();
            if (slot + add > this.max_locals) {
                this.max_locals = slot + add;
            }
            LocalVariableGen l = new LocalVariableGen(slot, name, type, start, end, orig_index);
            int i;
            if ((i = this.variable_vec.indexOf(l)) >= 0) {
                this.variable_vec.set(i, l);
                return l;
            }
            this.variable_vec.add(l);
            return l;
        }
        else {
            throw new IllegalArgumentException("Can not use " + type + " as type for local variable");
        }
    }
    
    public LocalVariableGen addLocalVariable(String name, Type type, int slot, InstructionHandle start, InstructionHandle end) {
        return this.addLocalVariable(name, type, slot, start, end, slot);
    }
    
    public LocalVariableGen addLocalVariable(String name, Type type, InstructionHandle start, InstructionHandle end) {
        return this.addLocalVariable(name, type, this.max_locals, start, end);
    }
    
    public void removeLocalVariable(LocalVariableGen l) {
        this.variable_vec.remove(l);
    }
    
    public void removeLocalVariables() {
        this.variable_vec.clear();
    }
    
    public LocalVariableGen[] getLocalVariables() {
        int size = this.variable_vec.size();
        LocalVariableGen[] lg = new LocalVariableGen[size];
        this.variable_vec.toArray(lg);
        for (int i = 0; i < size; ++i) {
            if (lg[i].getStart() == null && this.il != null) {
                lg[i].setStart(this.il.getStart());
            }
            if (lg[i].getEnd() == null && this.il != null) {
                lg[i].setEnd(this.il.getEnd());
            }
        }
        if (size > 1) {
            Arrays.sort(lg, new Comparator<LocalVariableGen>() {
                @Override
                public int compare(LocalVariableGen o1, LocalVariableGen o2) {
                    return o1.getIndex() - o2.getIndex();
                }
            });
            return lg;
        }
        return lg;
    }
    
    public LocalVariableTable getLocalVariableTable(ConstantPoolGen cp) {
        LocalVariableGen[] lg = this.getLocalVariables();
        int size = lg.length;
        LocalVariable[] lv = new LocalVariable[size];
        for (int i = 0; i < size; ++i) {
            lv[i] = lg[i].getLocalVariable(cp);
        }
        return new LocalVariableTable(cp.addUtf8("LocalVariableTable"), 2 + lv.length * 10, lv, cp.getConstantPool());
    }
    
    public LocalVariableTypeTable getLocalVariableTypeTable() {
        return this.local_variable_type_table;
    }
    
    public LineNumberGen addLineNumber(InstructionHandle ih, int src_line) {
        LineNumberGen l = new LineNumberGen(ih, src_line);
        this.line_number_vec.add(l);
        return l;
    }
    
    public void removeLineNumber(LineNumberGen l) {
        this.line_number_vec.remove(l);
    }
    
    public void removeLineNumbers() {
        this.line_number_vec.clear();
    }
    
    public LineNumberGen[] getLineNumbers() {
        LineNumberGen[] lg = new LineNumberGen[this.line_number_vec.size()];
        this.line_number_vec.toArray(lg);
        return lg;
    }
    
    public LineNumberTable getLineNumberTable(ConstantPoolGen cp) {
        int size = this.line_number_vec.size();
        LineNumber[] ln = new LineNumber[size];
        for (int i = 0; i < size; ++i) {
            ln[i] = ((LineNumberGen)this.line_number_vec.get(i)).getLineNumber();
        }
        return new LineNumberTable(cp.addUtf8("LineNumberTable"), 2 + ln.length * 4, ln, cp.getConstantPool());
    }
    
    public CodeExceptionGen addExceptionHandler(InstructionHandle start_pc, InstructionHandle end_pc, InstructionHandle handler_pc, ObjectType catch_type) {
        if (start_pc == null || end_pc == null || handler_pc == null) {
            throw new ClassGenException("Exception handler target is null instruction");
        }
        CodeExceptionGen c = new CodeExceptionGen(start_pc, end_pc, handler_pc, catch_type);
        this.exception_vec.add(c);
        return c;
    }
    
    public void removeExceptionHandler(CodeExceptionGen c) {
        this.exception_vec.remove(c);
    }
    
    public void removeExceptionHandlers() {
        this.exception_vec.clear();
    }
    
    public CodeExceptionGen[] getExceptionHandlers() {
        CodeExceptionGen[] cg = new CodeExceptionGen[this.exception_vec.size()];
        this.exception_vec.toArray(cg);
        return cg;
    }
    
    private CodeException[] getCodeExceptions() {
        int size = this.exception_vec.size();
        CodeException[] c_exc = new CodeException[size];
        for (int i = 0; i < size; ++i) {
            CodeExceptionGen c = (CodeExceptionGen)this.exception_vec.get(i);
            c_exc[i] = c.getCodeException(super.getConstantPool());
        }
        return c_exc;
    }
    
    public void addException(String class_name) {
        this.throws_vec.add(class_name);
    }
    
    public void removeException(String c) {
        this.throws_vec.remove(c);
    }
    
    public void removeExceptions() {
        this.throws_vec.clear();
    }
    
    public String[] getExceptions() {
        String[] e = new String[this.throws_vec.size()];
        this.throws_vec.toArray(e);
        return e;
    }
    
    private ExceptionTable getExceptionTable(ConstantPoolGen cp) {
        int size = this.throws_vec.size();
        int[] ex = new int[size];
        for (int i = 0; i < size; ++i) {
            ex[i] = cp.addClass((String)this.throws_vec.get(i));
        }
        return new ExceptionTable(cp.addUtf8("Exceptions"), 2 + 2 * size, ex, cp.getConstantPool());
    }
    
    public void addCodeAttribute(Attribute a) {
        this.code_attrs_vec.add(a);
    }
    
    public void removeLocalVariableTypeTable() {
        this.local_variable_type_table = null;
    }
    
    public void removeCodeAttribute(Attribute a) {
        this.code_attrs_vec.remove(a);
    }
    
    public void removeCodeAttributes() {
        this.local_variable_type_table = null;
        this.code_attrs_vec.clear();
    }
    
    public Attribute[] getCodeAttributes() {
        Attribute[] attributes = new Attribute[this.code_attrs_vec.size()];
        this.code_attrs_vec.toArray(attributes);
        return attributes;
    }
    
    public void addAnnotationsAsAttribute(ConstantPoolGen cp) {
        Attribute[] attrs;
        for (Attribute attr : attrs = AnnotationEntryGen.getAnnotationAttributes(cp, super.getAnnotationEntries())) {
            this.addAttribute(attr);
        }
    }
    
    public void addParameterAnnotationsAsAttribute(ConstantPoolGen cp) {
        if (!this.hasParameterAnnotations) {
            return;
        }
        Attribute[] attrs = AnnotationEntryGen.getParameterAnnotationAttributes(cp, this.param_annotations);
        if (attrs != null) {
            for (Attribute attr : attrs) {
                this.addAttribute(attr);
            }
        }
        return;
    }
    
    public Method getMethod() {
        String signature = this.getSignature();
        ConstantPoolGen _cp = super.getConstantPool();
        int name_index = _cp.addUtf8(super.getName());
        int signature_index = _cp.addUtf8(signature);
        byte[] byte_code = null;
        if (this.il != null) {
            byte_code = this.il.getByteCode();
        }
        LineNumberTable lnt = null;
        LocalVariableTable lvt = null;
        if (this.variable_vec.size() > 0 && !this.strip_attributes) {
            this.updateLocalVariableTable(this.getLocalVariableTable(_cp));
            this.addCodeAttribute(lvt = this.getLocalVariableTable(_cp));
        }
        if (this.local_variable_type_table != null) {
            if (lvt != null) {
                this.adjustLocalVariableTypeTable(lvt);
            }
            this.addCodeAttribute(this.local_variable_type_table);
        }
        if (this.line_number_vec.size() > 0 && !this.strip_attributes) {
            this.addCodeAttribute(lnt = this.getLineNumberTable(_cp));
        }
        Attribute[] code_attrs = this.getCodeAttributes();
        int attrs_len = 0;
        for (Attribute code_attr : code_attrs) {
            attrs_len += code_attr.getLength() + 6;
        }
        CodeException[] c_exc = this.getCodeExceptions();
        int exc_len = c_exc.length * 8;
        Code code = null;
        if (this.il != null && !this.isAbstract() && !this.isNative()) {
            Attribute[] attributes;
            for (Attribute a : attributes = this.getAttributes()) {
                if ((a instanceof Code)) {
                    this.removeAttribute(a);
                }
            }
            code = new Code(_cp.addUtf8("Code"), 8 + byte_code.length + 2 + exc_len + 2 + attrs_len, this.max_stack, this.max_locals, byte_code, c_exc, code_attrs, _cp.getConstantPool());
            this.addAttribute(code);
        }
        this.addAnnotationsAsAttribute(_cp);
        this.addParameterAnnotationsAsAttribute(_cp);
        ExceptionTable et = null;
        if (this.throws_vec.size() > 0) {
            this.addAttribute(et = this.getExceptionTable(_cp));
        }
        Method m = new Method(super.getAccessFlags(), name_index, signature_index, this.getAttributes(), _cp.getConstantPool());
        if (lvt != null) {
            this.removeCodeAttribute(lvt);
        }
        if (this.local_variable_type_table != null) {
            this.removeCodeAttribute(this.local_variable_type_table);
        }
        if (lnt != null) {
            this.removeCodeAttribute(lnt);
        }
        if (code != null) {
            this.removeAttribute(code);
        }
        if (et != null) {
            this.removeAttribute(et);
            return m;
        }
        return m;
    }
    
    private void updateLocalVariableTable(LocalVariableTable a) {
        LocalVariable[] lv = a.getLocalVariableTable();
        this.removeLocalVariables();
        for (LocalVariable l : lv) {
            InstructionHandle start = this.il.findHandle(l.getStartPC());
            InstructionHandle end = this.il.findHandle(l.getStartPC() + l.getLength());
            if (null == start) {
                start = this.il.getStart();
            }
            this.addLocalVariable(l.getName(), Type.getType(l.getSignature()), l.getIndex(), start, end, l.getOrigIndex());
        }
    }
    
    private void adjustLocalVariableTypeTable(LocalVariableTable lvt) {
        LocalVariable[] lv = lvt.getLocalVariableTable();
        LocalVariable[] lvg;
        for (LocalVariable element : lvg = this.local_variable_type_table.getLocalVariableTypeTable()) {
            int length2 = lv.length;
            int j = 0;
            while (j < length2) {
                LocalVariable l = lv[j];
                if (element.getName().equals(l.getName()) && element.getIndex() == l.getOrigIndex()) {
                    element.setLength(l.getLength());
                    element.setStartPC(l.getStartPC());
                    element.setIndex(l.getIndex());
                    break;
                }
                ++j;
            }
        }
    }
    
    public void removeNOPs() {
        if (this.il != null) {
            InstructionHandle next;
            for (InstructionHandle ih = this.il.getStart(); ih != null; ih = next) {
                next = ih.getNext();
                if (next != null && ih.getInstruction() instanceof NOP) {
                    try {
                        this.il.delete(ih);
                    }
                    catch (TargetLostException e) {
                        for (InstructionHandle target : e.getTargets()) {
                            for (InstructionTargeter targeter : target.getTargeters()) {
                                targeter.updateTarget(target, next);
                            }
                        }
                    }
                }
            }
        }
    }
    
    public void setMaxLocals(int m) {
        this.max_locals = m;
    }
    
    public int getMaxLocals() {
        return this.max_locals;
    }
    
    public void setMaxStack(int m) {
        this.max_stack = m;
    }
    
    public int getMaxStack() {
        return this.max_stack;
    }
    
    public String getClassName() {
        return this.class_name;
    }
    
    public void setClassName(String class_name) {
        this.class_name = class_name;
    }
    
    public void setReturnType(Type return_type) {
        this.setType(return_type);
    }
    
    public Type getReturnType() {
        return this.getType();
    }
    
    public void setArgumentTypes(Type[] arg_types) {
        this.arg_types = arg_types;
    }
    
    public Type[] getArgumentTypes() {
        return (Type[])this.arg_types.clone();
    }
    
    public void setArgumentType(int i, Type type) {
        this.arg_types[i] = type;
    }
    
    public Type getArgumentType(int i) {
        return this.arg_types[i];
    }
    
    public void setArgumentNames(String[] arg_names) {
        this.arg_names = arg_names;
    }
    
    public String[] getArgumentNames() {
        return (String[])this.arg_names.clone();
    }
    
    public void setArgumentName(int i, String name) {
        this.arg_names[i] = name;
    }
    
    public String getArgumentName(int i) {
        return this.arg_names[i];
    }
    
    public InstructionList getInstructionList() {
        return this.il;
    }
    
    public void setInstructionList(InstructionList il) {
        this.il = il;
    }
    
    @Override
    public String getSignature() {
        return Type.getMethodSignature(super.getType(), this.arg_types);
    }
    
    public void setMaxStack() {
        if (this.il != null) {
            this.max_stack = getMaxStack(super.getConstantPool(), this.il, this.getExceptionHandlers());
        }
        else {
            this.max_stack = 0;
        }
    }
    
    public void setMaxLocals() {
        if (this.il != null) {
            int max = this.isStatic() ? 0 : 1;
            if (this.arg_types != null) {
                for (Type arg_type : this.arg_types) {
                    max += arg_type.getSize();
                }
            }
            for (InstructionHandle ih = this.il.getStart(); ih != null; ih = ih.getNext()) {
                Instruction ins = ih.getInstruction();
                if (ins instanceof LocalVariableInstruction || ins instanceof RET || ins instanceof IINC) {
                    int index = ((IndexedInstruction)ins).getIndex() + ((TypedInstruction)ins).getType(super.getConstantPool()).getSize();
                    if (index > max) {
                        max = index;
                    }
                }
            }
            this.max_locals = max;
        }
        else {
            this.max_locals = 0;
        }
    }
    
    public void stripAttributes(boolean flag) {
        this.strip_attributes = flag;
    }
    
    public static int getMaxStack(ConstantPoolGen cp, InstructionList il, CodeExceptionGen[] et) {
        MethodGen.BranchStack branchTargets = new MethodGen.BranchStack();
        for (CodeExceptionGen element : et) {
            InstructionHandle handler_pc = element.getHandlerPC();
            if (handler_pc != null) {
                branchTargets.push(handler_pc, 1);
            }
        }
        int stackDepth = 0;
        int maxStackDepth = 0;
        InstructionHandle ih = il.getStart();
        while (ih != null) {
            Instruction instruction = ih.getInstruction();
            short opcode = instruction.getOpcode();
            int delta = instruction.produceStack(cp) - instruction.consumeStack(cp);
            stackDepth += delta;
            if (stackDepth > maxStackDepth) {
                maxStackDepth = stackDepth;
            }
            if ((instruction instanceof BranchInstruction)) {
                BranchInstruction branch = (BranchInstruction)instruction;
                if ((instruction instanceof Select)) {
                    Select select = (Select)branch;
                    InstructionHandle[] targets;
                    for (InstructionHandle target : targets = select.getTargets()) {
                        branchTargets.push(target, stackDepth);
                    }
                    ih = null;
                }
                else if (!(branch instanceof IfInstruction)) {
                    if (opcode == 168 || opcode == 201) {
                        branchTargets.push(ih.getNext(), stackDepth - 1);
                    }
                    ih = null;
                }
                branchTargets.push(branch.getTarget(), stackDepth);
            }
            else if (opcode == 191 || opcode == 169 || (opcode >= 172 && opcode <= 177)) {
                ih = null;
            }
            if (ih != null) {
                ih = ih.getNext();
            }
            if (ih == null) {
                MethodGen.BranchTarget bt = branchTargets.pop();
                if (bt != null) {
                    ih = bt.target;
                    stackDepth = bt.stackDepth;
                }
                else {
                    continue;
                }
            }
            else {
                continue;
            }
        }
        return maxStackDepth;
    }
    
    public void addObserver(MethodObserver o) {
        if (this.observers == null) {
            this.observers = new ArrayList();
        }
        this.observers.add(o);
    }
    
    public void removeObserver(MethodObserver o) {
        if (this.observers != null) {
            this.observers.remove(o);
        }
    }
    
    public void update() {
        if (this.observers != null) {
            for (MethodObserver observer : this.observers) {
                observer.notify(this);
            }
        }
    }
    
    @Override
    public final String toString() {
        String access = Utility.accessToString(super.getAccessFlags());
        String signature = Type.getMethodSignature(super.getType(), this.arg_types);
        signature = Utility.methodSignatureToString(signature, super.getName(), access, true, this.getLocalVariableTable(super.getConstantPool()));
        StringBuilder buf = new StringBuilder(signature);
        for (Attribute a : this.getAttributes()) {
            if (!(a instanceof Code) && !(a instanceof ExceptionTable)) {
                buf.append(" [").append(a).append("]");
            }
        }
        if (this.throws_vec.size() > 0) {
            for (String throwsDescriptor : this.throws_vec) {
                buf.append("\n\t\tthrows ").append(throwsDescriptor);
            }
        }
        return buf.toString();
    }
    
    public MethodGen copy(String class_name, ConstantPoolGen cp) {
        Method m = ((MethodGen)this.clone()).getMethod();
        MethodGen mg = new MethodGen(m, class_name, super.getConstantPool());
        if (super.getConstantPool() != cp) {
            mg.setConstantPool(cp);
            mg.getInstructionList().replaceConstantPool(super.getConstantPool(), cp);
            return mg;
        }
        return mg;
    }
    
    public List<AnnotationEntryGen> getAnnotationsOnParameter(int i) {
        this.ensureExistingParameterAnnotationsUnpacked();
        if (!this.hasParameterAnnotations || i > this.arg_types.length) {
            return null;
        }
        return this.param_annotations[i];
    }
    
    private void ensureExistingParameterAnnotationsUnpacked() {
        if (this.haveUnpackedParameterAnnotations) {
            return;
        }
        Attribute[] attrs = this.getAttributes();
        ParameterAnnotations paramAnnVisAttr = null;
        ParameterAnnotations paramAnnInvisAttr = null;
        for (Attribute attribute : attrs) {
            if ((attribute instanceof ParameterAnnotations)) {
                if (!this.hasParameterAnnotations) {
                    List<AnnotationEntryGen>[] parmList = new List[this.arg_types.length];
                    this.param_annotations = parmList;
                    for (int j = 0; j < this.arg_types.length; ++j) {
                        this.param_annotations[j] = new ArrayList();
                    }
                }
                this.hasParameterAnnotations = true;
                ParameterAnnotations rpa = (ParameterAnnotations)attribute;
                if ((rpa instanceof RuntimeVisibleParameterAnnotations)) {
                    paramAnnVisAttr = rpa;
                }
                else {
                    paramAnnInvisAttr = rpa;
                }
                ParameterAnnotationEntry[] parameterAnnotationEntries = rpa.getParameterAnnotationEntries();
                for (int i = 0; i < parameterAnnotationEntries.length; ++i) {
                    ParameterAnnotationEntry immutableArray = rpa.getParameterAnnotationEntries()[i];
                    List<AnnotationEntryGen> mutable = this.makeMutableVersion(immutableArray.getAnnotationEntries());
                    this.param_annotations[i].addAll(mutable);
                }
            }
        }
        if (paramAnnVisAttr != null) {
            this.removeAttribute(paramAnnVisAttr);
        }
        if (paramAnnInvisAttr != null) {
            this.removeAttribute(paramAnnInvisAttr);
        }
        this.haveUnpackedParameterAnnotations = true;
        return;
    }
    
    private List<AnnotationEntryGen> makeMutableVersion(AnnotationEntry[] mutableArray) {
        List<AnnotationEntryGen> result = new ArrayList();
        for (AnnotationEntry element : mutableArray) {
            result.add(new AnnotationEntryGen(element, this.getConstantPool(), false));
        }
        return result;
    }
    
    public void addParameterAnnotation(int parameterIndex, AnnotationEntryGen annotation) {
        this.ensureExistingParameterAnnotationsUnpacked();
        if (!this.hasParameterAnnotations) {
            List<AnnotationEntryGen>[] parmList = new List[this.arg_types.length];
            this.param_annotations = parmList;
            this.hasParameterAnnotations = true;
        }
        List<AnnotationEntryGen> existingAnnotations = this.param_annotations[parameterIndex];
        if (existingAnnotations != null) {
            existingAnnotations.add(annotation);
        }
        else {
            List<AnnotationEntryGen> l = new ArrayList();
            l.add(annotation);
            this.param_annotations[parameterIndex] = l;
        }
    }
    
    public static BCELComparator getComparator() {
        return MethodGen.bcelComparator;
    }
    
    public static void setComparator(BCELComparator comparator) {
        MethodGen.bcelComparator = comparator;
    }
    
    @Override
    public boolean equals(Object obj) {
        return MethodGen.bcelComparator.equals(this, obj);
    }
    
    @Override
    public int hashCode() {
        return MethodGen.bcelComparator.hashCode(this);
    }
    
    static {
        MethodGen.bcelComparator = new BCELComparator() {
            @Override
            public boolean equals(Object o1, Object o2) {
                MethodGen THIS = (MethodGen)o1;
                MethodGen THAT = (MethodGen)o2;
                return THIS.getName().equals(THAT.getName()) && THIS.getSignature().equals(THAT.getSignature());
            }
            
            @Override
            public int hashCode(Object o) {
                MethodGen THIS = (MethodGen)o;
                return THIS.getSignature().hashCode() ^ THIS.getName().hashCode();
            }
        };
    }
    
    static final class BranchTarget
    {
        final InstructionHandle target;
        final int stackDepth;
        
        BranchTarget(InstructionHandle target, int stackDepth) {
            this.target = target;
            this.stackDepth = stackDepth;
        }
    }
    
    static final class BranchStack
    {
        private final Stack<MethodGen.BranchTarget> branchTargets;
        private final Map<InstructionHandle, MethodGen.BranchTarget> visitedTargets;
        
        BranchStack() {
            this.branchTargets = new Stack();
            this.visitedTargets = new HashMap();
        }
        
        public void push(InstructionHandle target, int stackDepth) {
            if (this.visited(target)) {
                return;
            }
            this.branchTargets.push(this.visit(target, stackDepth));
            return;
        }
        
        public MethodGen.BranchTarget pop() {
            if (!this.branchTargets.empty()) {
                MethodGen.BranchTarget bt = (MethodGen.BranchTarget)this.branchTargets.pop();
                return bt;
            }
            return null;
        }
        
        private MethodGen.BranchTarget visit(InstructionHandle target, int stackDepth) {
            MethodGen.BranchTarget bt = new MethodGen.BranchTarget(target, stackDepth);
            this.visitedTargets.put(target, bt);
            return bt;
        }
        
        private boolean visited(InstructionHandle target) {
            return this.visitedTargets.get(target) != null;
        }
    }
}
